#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
import sys, os
import optparse
from StringIO import StringIO

import ZODB
from ZODB.FileStorage import FileStorage

from schooltool.db_exporter import *
from schooltool.db_exporter.xls import XLSWriter


DB_SCHEMA_MODULE = 'schooltool.db_exporter.mock_0_11_4'



def mock_db_schema(schema_module_name):
    from schooltool.db_exporter.mock import ModulesSnapshot
    snapshot = ModulesSnapshot()
    _temp = __import__(schema_module_name, globals(), locals(), ['schema'], 0)
    snapshot.mock(_temp.schema)
    return snapshot


def load_sheet_writers(module_name):
    _temp = __import__(module_name, globals(), locals(), ['sheet_writers'], 0)
    return _temp.sheet_writers


def check_db_version(module_name, db_root, verbose=False):
    _temp = __import__(module_name, globals(), locals(),
                       ['check_db_version'], 0)
    return _temp.check_db_version(db_root, verbose=verbose)


def open_database(filename, read_only=True):
    print 'Opening database', filename
    try:
        if not os.access(filename, os.F_OK):
            print "Can't open the file."
            return None, None
        if not os.access('%s.index' % filename, os.F_OK):
            print ('Index file does not exist - '
                   'opening this database may take a while...')
            if read_only:
                print ('Pass --repair-index option to speed up '
                       'database access later on.')
            else:
                print 'The index will be repaired.'
        print ''
        storage = FileStorage(filename, read_only=read_only)
        db = ZODB.DB(storage)
        connection = db.open()
    except ZODB.POSException.POSError:
        # Hide exceptions.  They're not helpful -
        # ZODB.POSException.ReadOnlyError just doesn't quite say
        # 'file not found'
        print 'Failed to open the database.'
        return None, None
    return db, connection


def iter_storage(connection):
    storage = connection._storage
    next_oid = None
    while True:
        oid, tid, data, next_oid = storage.record_iternext(next_oid)
        yield connection.get(oid)
        if next_oid is None:
            return


def export(connection, output_filename, verbose=False, force=False):

    print 'Configuring data model.'

    snapshot = mock_db_schema(DB_SCHEMA_MODULE)
    print 'Configuring XLS exporter.'
    sheet_writers = load_sheet_writers(DB_SCHEMA_MODULE)
    exporter = XLSWriter(sheet_writers.values())
    exporter.begin()
    print ''

    ZODB.broken.broken_cache.clear()

    root = connection.root()

    if not check_db_version(DB_SCHEMA_MODULE, root, verbose=True):
        print ''
        print "Don't how to read this database properly."
        if not force:
            print 'Aborting (you can try --force option)'
            return
        print 'Trying anyway.'
    print ''

    seen_broken = {}
    n_accepted = 0
    for n, obj in enumerate(iter_storage(connection)):
        if isinstance(obj, ZODB.broken.Broken):
            name = '%s.%s' % (obj.__class__.__module__,
                              obj.__class__.__name__)
            if name not in seen_broken:
                seen_broken[name] = 1
            else:
                seen_broken[name] += 1
        else:
            if exporter.read(obj):
                n_accepted += 1

        if n % 500 == 499:
            sys.stdout.write('.')
            sys.stdout.flush()

    print '\n'
    print 'Objects in database:', n+1
    print 'Objects accepted for export:', n_accepted
    print ''

    print 'Exporting to', os.path.abspath(output_filename)
    exporter.write(output_filename)

    snapshot.restore()

    print 'Done.'

    if verbose:
        print ''
        print 'Broken (unknown) objects seen:', sum(seen_broken.values())
        print 'Broken classes:'
        for module, cls in sorted(ZODB.broken.broken_cache):
            name = '%s.%s' % (module, cls)
            print '  %s (%s)' % (name, seen_broken.get(name, '?'))


def parse_args():
    """Parse the command line arguments"""
    parser = optparse.OptionParser(
        usage="usage: %prog [Data.fs filename] [options]")

    parser.add_option("-o",
                      dest="output_filename",
                      default="export.xls",
                      help="Specify output .xls file.")
    parser.add_option("--repair-index",
                      action="store_true", dest="repair_index", default=False,
                      help="""
                      Repair database index if needed.
                      This may create additional files.
                      """.strip())
    parser.add_option("--force",
                      action="store_true", dest="force", default=False,
                      help="Export even if database versions are not right")
    parser.add_option("-v", "--verbose",
                      action="store_true", dest="verbose", default=False,
                      help="More detailed output.")
    options, args = parser.parse_args()
    return options, args


def run():
    options, args = parse_args()
    if len(args) > 0:
        dbname = os.path.abspath(args[0])
    else:
        dbname = 'Data.fs'

    db, connection = open_database(
        dbname, read_only=not options.repair_index)

    if connection is not None:
        try:
            export(connection, options.output_filename,
                   verbose=options.verbose, force=options.force)
        finally:
            connection.close()
            db.close()




