#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import datetime

from persistent import Persistent
from persistent.dict import PersistentDict

from schooltool.db_exporter.relationship import URIObject
from schooltool.db_exporter.relationship import RelationshipPropertyBase
from schooltool.db_exporter.relationship import LinkSet, Link
from schooltool.db_exporter.xls import SimpleSheetWriter
from schooltool.db_exporter.xls import WARNING, ERROR


RELATIONSHIP_KEY = 'schoolbell.relationship.RelationshipLinks'

class RelationshipProperty(RelationshipPropertyBase):
    annotation_key = RELATIONSHIP_KEY


def iter_date_range(first, last):
    current = first
    while current <= last:
        yield current
        current += datetime.date.resolution


def zope_name(obj):
    return obj.__name__


class ContainedObject(Persistent):
    __name__ = None
    __parent__ = None


class SimpleObject(ContainedObject):
    title = None
    description = None


URIMembership = URIObject('http://schooltool.org/ns/membership',
                          'Membership', 'The membership relationship.')

URIGroup = URIObject('http://schooltool.org/ns/membership/group',
                     'Group', 'A role of a containing group.')

URIMember = URIObject('http://schooltool.org/ns/membership/member',
                      'Member', 'A group member role.')


class Person(ContainedObject):
    photo = None
    username = None
    _hashed_password = None
    title = None
    username = None
    groups = RelationshipProperty(URIMembership, URIMember, URIGroup)
    __annotations__ = None
    @property
    def _details(self):
        key = 'schoolbell.app.PersonDetails'
        if key not in self.__annotations__:
            return PersonDetails()
        return self.__annotations__[key]


class PersonDetails(ContainedObject):
    nickname = None
    primary_email = None
    secondary_email = None
    primary_phone = None
    secondary_phone = None
    mailing_address = None
    home_page = None


class PersonSheetWriter(SimpleSheetWriter):

    sheet_name = 'Persons'
    collect_instances = Person

    def write(self):
        persons = sorted(self.objects, key=zope_name)

        fields = (
            'User Name', 'Prefix',
            'First Name', 'Middle Name', 'Last Name',
            'Suffix', 'Preferred Name', 'Birth Date', 'Gender',
            'Password', 'ID', 'Ethnicity', 'Language',
            'Place of birth', 'Citizenship' )

        self.writeRow(
            fields,
            style=self.styles.header)

        for person in persons:
            names = person.title.split(' ')
            if len(names) >=2:
                first_name = ' '.join(names[:-1])
                last_name = names[-1]
            else:
                first_name = person.title
                last_name = ''

            email = (person._details.primary_email or
                     person._details.secondary_email)

            phone = (person._details.primary_phone or
                     person._details.secondary_phone)

            self.writeRow([person.username, ''])

            self.appendRow(
                [first_name],
                style=(first_name and self.styles.default or self.styles.error))
            self.appendRow(['']) # middle name
            self.appendRow(
                [last_name],
                style=(last_name and self.styles.default or self.styles.error))
            if not first_name or not last_name:
                self.exporter.error(
                    WARNING,
                    "User %s (%r) missing first name or last name" % (
                        person.username, person.title))

            self.appendRow([
                '', person._details.nickname, None, None,
                None, None, None, None,
                None, None,
                ])


class TimetableDict(PersistentDict):
    __name__ = 'timetables'
    __parent__ = None


class Term(ContainedObject):
    first = None
    last = None
    title = None
    _schooldays = None


class SchoolYearSheetWriter(SimpleSheetWriter):

    sheet_name = 'School Years'
    collect_instances = Term
    schoolyear_id = None

    def write(self):
        if not self.objects:
            return
        first = min([t.first for t in self.objects])
        last = max([t.last for t in self.objects])

        if first.year == last.year:
            self.schoolyear_id = '%s' % first.year
        else:
            self.schoolyear_id = '%s-%s' % (first.year, last.year)

        self.writeRow(
            ['ID', 'Title', 'Start', 'End'],
            style=self.styles.header)

        self.writeRow(
            [self.schoolyear_id, self.schoolyear_id])

        self.appendRow(
            [first, last],
            style=self.styles.date)


class WithSchoolyearSheetWriter(SimpleSheetWriter):

    def __init__(self, *args, **kw):
        SimpleSheetWriter.__init__(self, *args, **kw)
        self.requires_sheets = self.requires_sheets + ('School Years', )

    @property
    def schoolyear_id(self):
        year_writer = self.exporter.getWriter('School Years')
        return year_writer.schoolyear_id


class TermSheetWriter(WithSchoolyearSheetWriter):

    sheet_name = 'Terms'
    collect_instances = Term
    is_weekend = None

    def __init__(self, *args, **kw):
        WithSchoolyearSheetWriter.__init__(self, *args, **kw)
        self.is_weekend = dict([(n, False) for n in range(1, 8)])

    def write(self):
        self.guessWeekends()

        self.writeTerms()
        self.writeHolidays()
        self.writeWeekends()
        self.writeWorkingWeekends()

    def guessWeekends(self):
        total_weekdays = dict([(n, 0) for n in range(1, 8)])
        schoolday_weekdays = dict([(n, 0) for n in range(1, 8)])
        for term in self.objects:
            for day in iter_date_range(term.first, term.last):
                weekday = day.isoweekday()
                total_weekdays[weekday] += 1
                if day in term._schooldays:
                    schoolday_weekdays[weekday] += 1
        # ratio of schooldays to a weekday
        # for it to be considered not a weekend
        MIN_SCHOOLDAY_RATIO = 0.3
        for n in range(1, 8):
            n_days = float(max(1, total_weekdays[n]))
            n_schooldays = float(max(1, schoolday_weekdays[n]))
            self.is_weekend[n] = bool(n_schooldays/n_days < (1-MIN_SCHOOLDAY_RATIO))

    def writeTerms(self):
        self.writeRow(
            ['School Year', 'ID', 'Title', 'Start', 'End'],
            style=self.styles.header)
        terms = sorted(self.objects, key=lambda t: t.first)
        for term in terms:
            self.writeRow(
                [self.schoolyear_id,
                 term.__name__ or '', term.title or ''])
            self.appendRow(
                [term.first, term.last],
                style=self.styles.date)
        self.skipRow()

    def writeHolidays(self):
        holidays = set()
        for term in self.objects:
            for day in iter_date_range(term.first, term.last):
                if (not self.is_weekend[day.isoweekday()]
                    and day not in term._schooldays):
                    holidays.add(day)
        if not holidays:
            return
        self.writeRow(
            ['Holidays'], style=self.styles.header)
        for day in sorted(holidays):
            self.writeRow([day, day], style=self.styles.date)
        self.skipRow()

    def writeWorkingWeekends(self):
        exceptions = set()
        for term in self.objects:
            for day in iter_date_range(term.first, term.last):
                if (self.is_weekend[day.isoweekday()]
                    and day in term._schooldays):
                    exceptions.add(day)
        if not exceptions:
            return
        self.writeRow(
            ['Working weekends'], style=self.styles.header)
        for day in sorted(exceptions):
            self.writeRow([day], style=self.styles.date)
        self.skipRow()

    def writeWeekends(self):
        self.writeRow(
            ['Weekends'], style=self.styles.header)
        self.writeRow(
            ['Monday', 'Tuesday', 'Wednesday', 'Thursday',
             'Friday', 'Saturday', 'Sunday'])
        self.writeRow([w and 'X' or None for w in self.is_weekend.values()])
        self.skipRow()


URICourseSections = URIObject('http://schooltool.org/ns/coursesections',
                          'Course Sections',
                          'The sections that implement a course.')

URICourse = URIObject('http://schooltool.org/ns/coursesections/course',
                      'Course', 'A section that implements a course.')

URISectionOfCourse = URIObject(
                      'http://schooltool.org/ns/coursesections/section',
                      'Section', 'A course of study.')


class Course(SimpleObject):
    sections = RelationshipProperty(URICourseSections, URICourse,
                                    URISectionOfCourse)


class CourseSheetWriter(WithSchoolyearSheetWriter):

    sheet_name = 'Courses'
    collect_instances = Course

    def write(self):
        courses = sorted(self.objects, key=zope_name)

        self.writeRow(
            ['School Year', 'ID', 'Title', 'Description'],
            style=self.styles.header)

        for course in courses:
            self.writeRow(
                [self.schoolyear_id, course.__name__,
                 course.title, course.description])


def flatten_members(object):
    result = set()
    for member in object.members:
        if isinstance(member, Person):
            result.add(member)
        elif isinstance(member, Group):
            result.update(flatten_members(member))
    return list(result)


class Group(SimpleObject):
    members = RelationshipProperty(URIMembership, URIGroup, URIMember)


class GroupSheetWriter(WithSchoolyearSheetWriter):

    sheet_name = 'Groups'
    collect_instances = Group

    def write(self):
        groups = sorted(self.objects, key=zope_name)
        for group in groups:
            self.writeGroup(group)
            members = flatten_members(group)
            if members:
                self.appendRow(['']) # border
                self.writeMembers(members)

    def appendRow(self, cells, row=-1, style=None):
        if style is None:
            style = self.styles.right_border
        super(GroupSheetWriter, self).appendRow(cells, row=row, style=style)

    def writeGroup(self, group):
        self.writeRow(['Group Title'], style=self.styles.header_top_border)
        self.appendRow([group.title], style=self.styles.top_right_border)

        self.writeRow(['ID'], style=self.styles.header)
        self.appendRow([group.__name__])

        self.writeRow(['School Year'], style=self.styles.header)
        self.appendRow([self.schoolyear_id])

        self.writeRow(['Description'], style=self.styles.header)
        self.appendRow([group.description])
        self.skipRow()

    def writeMembers(self, members):
        self.writeRow(['Members'], style=self.styles.header)
        self.appendRow(['']) # border
        for member in sorted(members, key=zope_name):
            if isinstance(member, Person):
                self.writeRow([member.__name__])
                self.appendRow(['']) # border
        self.skipRow()


URIInstruction = URIObject('http://schooltool.org/ns/instruction',
                          'Instruction', 'The instruction relationship.')

URISection = URIObject('http://schooltool.org/ns/instruction/section',
                     'Section', 'A role of a containing section.')

URIInstructor = URIObject('http://schooltool.org/ns/instruction/instructor',
                      'Instructor', 'A section instructor role.')


class Section(SimpleObject):
    title = None
    description = None
    timetables = None

    instructors = RelationshipProperty(URIInstruction, URISection,
                                       URIInstructor)

    courses = RelationshipProperty(URICourseSections, URISectionOfCourse,
                                   URICourse)

    members = RelationshipProperty(URIMembership, URIGroup, URIMember)

    # Some helpers that were not in the actual object
    def expandForTerms(self):
        tt_keys = self.timetables.keys()
        term_keys = set([key.split('.')[0] for key in tt_keys])
        return [TermSection(self, key) for key in term_keys]


class TermSection(object):
    _original_section = None
    _term_id = None

    def __init__(self, section, term_id):
        self.__dict__['_original_section'] = section
        self.__dict__['_term_id'] = term_id

    def __getattr__(self, attr):
        return getattr(self.__dict__['_original_section'], attr)

    @property
    def term_id(self):
        return self.__dict__['_term_id']

    @property
    def __name__(self):
        return '%s_%s' % (
            self.__dict__['_term_id'],
            self.__dict__['_original_section'].__name__)

class SectionMultiSheetsWriter(WithSchoolyearSheetWriter):

    requires_sheets = ('Courses', 'Persons', 'Groups', 'Terms')
    # Just for reference, this sheet is never created
    sheet_name = 'Section sheets'
    collect_instances = Section
    sheet_writers = None

    def __init__(self, *args, **kw):
        WithSchoolyearSheetWriter.__init__(self, *args, **kw)

    def begin(self):
        self.sheet_writers = []

    @property
    def sections(self):
        expanded_sections = []
        for section in self.objects:
            expanded = section.expandForTerms()
            if not expanded:
                # Sigh, this one is not scheduled for a term.
                expanded_sections.append(section)
            else:
                expanded_sections.extend(expanded)
        return expanded_sections

    def getTerm(self, section):
        term_id = getattr(section, 'term_id', None)
        if term_id is None:
            return None
        term_writer = self.exporter.getWriter('Terms')
        for obj in term_writer.objects:
            if obj.__name__ == term_id:
                return obj
        return None

    def write(self):
        sections = sorted(self.sections, key=zope_name)
        by_term = {}
        for section in sections:
            term = self.getTerm(section)
            if term not in by_term:
                by_term[term] = []
            by_term[term].append(section)

        term_sort_key = lambda t: t is not None and t.first.isoformat() or None

        for term in sorted(by_term, key=term_sort_key):
            writer = SectionSheetWriter(self.exporter, term)
            self.sheet_writers.append(writer)
            writer.begin()
            for term_section in by_term[term]:
                writer.read(term_section)

        for writer in self.sheet_writers:
            writer.write()


class SectionSheetWriter(WithSchoolyearSheetWriter):

    term = None

    def __init__(self, exporter, term):
        WithSchoolyearSheetWriter.__init__(self, exporter)
        self.term = term

    def begin(self):
        if self.term is None:
            self.sheet_name = 'Sections with no term'
            self.exporter.error(
                WARNING,
                "There are sections with no terms assigned, "
                "putting them to sheet %r "
                "(they will fail to import)" % (
                    self.sheet_name))
        else:
            self.sheet_name = 'Sections %s %s' % (
                self.schoolyear_id, self.term.title)
        WithSchoolyearSheetWriter.begin(self)

    def appendRow(self, cells, row=-1, style=None):
        if style is None:
            style = self.styles.right_border
        super(SectionSheetWriter, self).appendRow(cells, row=row, style=style)

    def write(self):
        self.writeRow(['School Year'], style=self.styles.header)
        self.appendRow([self.schoolyear_id], style=self.styles.default)
        self.appendRow(['Term'], style=self.styles.header)
        if self.term is not None:
            self.appendRow([self.term.__name__], style=self.styles.default)
        else:
            self.appendRow([''], style=self.styles.error)
        self.skipRow()

        sections = sorted(self.objects, key=zope_name)
        for section in sections:
            self.writeSection(section)
        self.writeRow([' ']) # Workaround for our stupid importer

    def writeSection(self, section, term=None):
        self.writeRow(['Section Title'], style=self.styles.header_top_border)
        self.appendRow([section.title], style=self.styles.top_right_border)

        self.writeRow(['ID'], style=self.styles.header)
        self.appendRow([section.__name__])

        self.writeRow(['Description'], style=self.styles.header)
        self.appendRow([section.description])

        self.skipRow()

        courses = list(section.courses)
        if courses:
            self.appendRow(['']) # border
            self.writeObjects('Courses', section.courses)

        students = flatten_members(section)
        if students:
            self.appendRow(['']) # border
            self.writeObjects('Students', students)

        instructors = list(section.instructors)
        if instructors:
            self.appendRow(['']) # border
            self.writeObjects('Instructors', section.instructors)

    def writeObjects(self, header, objects):
        self.writeRow([header], style=self.styles.header)
        self.appendRow(['']) # border
        for obj in sorted(objects, key=zope_name):
            self.writeRow([obj.__name__])
            self.appendRow(['']) # border
        self.skipRow()


# 'schema' defines the database configuration
schema = {
    'schoolbell.relationship.relationship.LinkSet': LinkSet,
    'schoolbell.relationship.relationship.Link': Link,
    'schoolbell.relationship.uri.URIObject': URIObject,
    'schooltool.timetable.Term': Term,
    'schooltool.timetable.TimetableDict': TimetableDict,
    'schooltool.app.Person': Person,
    'schoolbell.app.Person': Person,
    'schoolbell.app.app.PersonDetails': PersonDetails,
    'schoolbell.app.Group': Group,
    'schooltool.app.Group': Group,
    'schooltool.app.Course': Course,
    'schooltool.app.Section': Section,
    }

# 'sheet_writers' use the objects from schema to build .xls
sheet_writers = {
    'terms': TermSheetWriter,
    'school_years': SchoolYearSheetWriter,
    'persons': PersonSheetWriter,
    'groups': GroupSheetWriter,
    'courses': CourseSheetWriter,
    'sections': SectionMultiSheetsWriter,
    }


def check_db_version(db_root, verbose=False):
    if verbose:
        print 'This module can read SchoolTool 0.11.3 and 0.11.4 databases.'
    versions = db_root.get('zope.app.generations', None)
    if versions is None:
        if verbose:
            print 'No database version information found.'
            return False
    if verbose:
        print 'Database versions:'
        for entry in sorted(versions):
            print '  %s v. %s' % (entry, versions[entry])

    can_evolve = True
    schoolbell_ver = versions.get('schoolbell.app', None)
    if schoolbell_ver is None:
        if verbose:
            print 'Warning: No version information for schoolbell.app'
        can_evolve = False

    schooltool_ver = versions.get('schooltool', None)
    if schooltool_ver is None:
        if verbose:
            print 'Warning: No version information for schooltool'
        can_evolve = False
    else:
        required_version = 3
        can_evolve = bool(schooltool_ver == required_version)
        if ((schooltool_ver != required_version) and verbose):
            print ('Warning: Schooltool database version %d, '
                   'required version is %d' %
                       (schooltool_ver, required_version))

    return can_evolve
