#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
import sys
import xlwt

from schooltool.db_exporter import TopologicalSort


ERROR = 'Error'
WARNING = 'Warning'


_short_name_domains = {}
def unique_short_name(name, domain='', maxlen=30):
    if domain not in _short_name_domains:
        _short_name_domains[domain] = {}
    name_map = _short_name_domains[domain]
    if name in name_map:
        return name_map[name]
    if (len(name) <= maxlen and
        name not in name_map.values()):
        name_map[name] = name
        return name
    i = 1
    while True:
        suffix = '~%d' % i
        if len(suffix) > maxlen:
            raise Exception("Can't make a short name (len %d) from %r" % (
                    maxlen, name))
        short_name = name[:maxlen-len(suffix)] + suffix
        if short_name not in name_map.values():
            name_map[name] = short_name
            return short_name
        i += 1


class CellStyleSet(object):

    def __init__(self):
        self._font_bold = xlwt.Font()
        self._font_bold.bold = True

        self.default = xlwt.XFStyle()
        self.header = xlwt.easyxf(
            'font: bold on; '
            'pattern: pattern solid, fore_colour yellow')
        self.error = xlwt.easyxf(
            'pattern: pattern solid, fore_colour red')
        self.date = xlwt.XFStyle()
        self.date.num_format_str = 'YYYY-MM-DD'
        self.header_top_border = xlwt.easyxf(
            'font: bold on; '
            'pattern: pattern solid, fore_colour yellow; '
            'borders: top thin')
        self.top_right_border = xlwt.easyxf(
            'borders: top thin, right thin')
        self.right_border = xlwt.easyxf(
            'borders: right thin')


class XLSWriter(object):

    sheet_writers = None
    errors = None
    workbook = None
    styles = None # cell styles

    def __init__(self, writer_factories):
        self.factories = writer_factories

    def begin(self):
        self.workbook = xlwt.Workbook()
        self.styles = CellStyleSet()
        self.errors = {}
        self.sheet_writers = {}
        for factory in self.factories:
            if factory.sheet_name in self.sheet_writers:
                raise Exception('Duplicate sheet %r writer %r (%r)' % (
                    factory.sheet_name, factory,
                    self.sheet_writers[factory.sheet_name]))
            self.sheet_writers[factory.sheet_name] = factory(self)
        sorter = TopologicalSort(dict(
            [(writer.sheet_name, writer.requires_sheets)
             for writer in self.sheet_writers.values()]))
        self.writer_order = sorter()
        print 'Sheet order:', self.writer_order
        for name in self.writer_order:
            writer = self.getWriter(name)
            writer.begin()

    def getWriter(self, sheet_name):
        return self.sheet_writers[sheet_name]

    def read(self, obj):
        accepted = False
        for name in self.writer_order:
            writer = self.getWriter(name)
            if writer.read(obj):
                accepted = True
        return accepted

    def write(self, filename):
        for name in self.writer_order:
            writer = self.getWriter(name)
            writer.write()
        datafile = file(filename, 'w')
        self.workbook.save(datafile)
        datafile.close()

    def error(self, group, message):
        if group not in self.errors:
            self.errors[group] = []
        self.errors[group].append(message)
        sys.stderr.write('%s: %s\n' % (group, message))


class SheetWriter(object):

    sheet_name = None # override in child classes
    requires_sheets = () # other sheets that need to be filled in before this one
    sheet = None
    exporter = None

    def __init__(self, exporter):
        self.exporter = exporter

    def begin(self):
        if self.sheet_name is not None:
            sheet_name = unique_short_name(
                self.sheet_name,
                domain='SheetWriter.sheet_name',
                maxlen=30)
            self.sheet = self.workbook.add_sheet(sheet_name)

    @property
    def styles(self):
        return self.exporter.styles

    @property
    def workbook(self):
        return self.exporter.workbook

    def read(self, obj):
        raise NotImplemented('...')

    def write(self):
        raise NotImplemented('...')

    @property
    def next_row(self):
        return self.sheet.rows and (max(self.sheet.rows) + 1) or 0

    @property
    def last_row(self):
        return self.sheet.rows and (max(self.sheet.rows)) or 0

    def writeRow(self, cells, row=-1, col=0, style=None):
        """Write cells to a (new) row."""
        if style is None:
            style = self.styles.default
        if row == -1:
            row = self.next_row
        for n, cell in enumerate(cells):
            if cell is None:
                continue
            self.sheet.write(row, col + n, cell, style)

    def appendRow(self, cells, row=-1, style=None):
        """Append cells to the (last) row."""
        if not self.sheet.rows:
            self.writeRow(cells, style=style)
        if row == -1:
            row = self.last_row
        # it will start on col 1 in empty rows, but oh well.
        col = self.sheet.row(row).get_max_col() + 1
        self.writeRow(cells, row, col, style=style)

    def skipRow(self):
        self.sheet.write(self.last_row + 1, 0, '', self.styles.default)


class SimpleSheetWriter(SheetWriter):
    """A sheet writer that simply collects instances."""

    objects = None
    collect_instances = object

    def __init__(self, *args, **kw):
        SheetWriter.__init__(self, *args, **kw)
        self.objects = set()

    def read(self, obj):
        if isinstance(obj, self.collect_instances):
            self.objects.add(obj)
            return True
        return False

